@@ -844,6 +844,33 @@ static int rans_cpu = 0xFFFF; // all
 #  define UNUSED
 #endif
 
+#if defined(__APPLE__)
+/*
+   MacOS before 12.2 (a.k.a. Darwin 21.3) had a bug that could cause random
+   failures on some AVX512 operations due to opmask registers not being restored
+   correctly following interrupts.  For simplicity, check that the major version
+   is 22 or higher before using AVX512.
+   See https://community.intel.com/t5/Software-Tuning-Performance/MacOS-Darwin-kernel-bug-clobbers-AVX-512-opmask-register-state/m-p/1327259
+*/
+
+#include <sys/utsname.h>
+static inline int not_ancient_darwin(void) {
+    static long version = 0;
+    if (!version) {
+        struct utsname uname_info;
+        if (uname(&uname_info) == 0) {
+            version = strtol(uname_info.release, NULL, 10);
+        }
+    }
+    return version >= 22;
+}
+#else
+static inline int not_ancient_darwin(void) {
+    return 1;
+}
+#endif
+
+
 // CPU detection is performed once.  NB this has an assumption that we're
 // not migrating between processes with different instruction stes, but
 // to date the only systems I know of that support this don't have different
@@ -862,6 +889,11 @@ static int is_amd       UNUSED = 0;
 #define HAVE_HTSCODECS_TLS_CPU_INIT
 static void htscodecs_tls_cpu_init(void) {
     unsigned int eax = 0, ebx = 0, ecx = 0, edx = 0;
+    unsigned int have_xsave UNUSED = 0;
+    unsigned int have_avx   UNUSED = 0;
+    uint64_t xcr0 UNUSED = 0ULL;
+    const uint64_t xcr0_can_use_avx UNUSED = (1ULL << 2);
+    const uint64_t xcr0_can_use_avx512 UNUSED = (7ULL << 5);
     // These may be unused, depending on HAVE_* config.h macros
 
     int level = __get_cpuid_max(0, NULL);
@@ -877,15 +909,43 @@ static void htscodecs_tls_cpu_init(void) {
 #endif
 #if defined(bit_SSE4_1)
         have_sse4_1 = ecx & bit_SSE4_1;
+#endif
+#if defined(bit_AVX)
+        have_avx = ecx & bit_AVX;
+#endif
+#if defined(bit_XSAVE) && defined(bit_OSXSAVE)
+        have_xsave = (ecx & bit_XSAVE) && (ecx & bit_OSXSAVE);
+        if (have_xsave) {
+            /* OSXSAVE tells us it's safe to use XGETBV to read XCR0
+               which then describes if AVX / AVX512 instructions can be
+               executed.  See Intel 64 and IA-32 Architectures Software
+               Developer’s Manual Vol. 1 sections 13.2 and 13.3.
+
+               Use inline assembly for XGETBV here to avoid problems
+               with builtins either not working correctly, or requiring
+               specific compiler options to be in use.  Also emit raw
+               bytes here as older toolchains may not have the XGETBV
+               instruction.
+            */
+            __asm__ volatile (".byte 0x0f, 0x01, 0xd0" :
+                              "=d" (edx), "=a" (eax) :
+                              "c" (0));
+            xcr0 = ((uint64_t) edx << 32) | eax;
+        }
 #endif
     }
-    if (level >= 7) {
+    // AVX2 and AVX512F depend on XSAVE, AVX and bit 2 of XCR0.
+    if (level >= 7 && have_xsave && have_avx
+        && (xcr0 & xcr0_can_use_avx) == xcr0_can_use_avx) {
         __cpuid_count(7, 0, eax, ebx, ecx, edx);
 #if defined(bit_AVX2)
         have_avx2 = ebx & bit_AVX2;
 #endif
 #if defined(bit_AVX512F)
-        have_avx512f = ebx & bit_AVX512F;
+        // AVX512 depends on bits 5:7 of XCR0
+        if ((xcr0 & xcr0_can_use_avx512) == xcr0_can_use_avx512
+            && not_ancient_darwin())
+            have_avx512f = ebx & bit_AVX512F;
 #endif
     }
 
@@ -900,14 +960,6 @@ unsigned char *(*rans_enc_func(int do_simd, int order))
      unsigned char *out,
      unsigned int *out_size) {
 
-    int have_e_sse4_1  = have_sse4_1;
-    int have_e_avx2    = have_avx2;
-    int have_e_avx512f = have_avx512f;
-
-    if (!(rans_cpu & RANS_CPU_ENC_AVX512)) have_e_avx512f = 0;
-    if (!(rans_cpu & RANS_CPU_ENC_AVX2))   have_e_avx2    = 0;
-    if (!(rans_cpu & RANS_CPU_ENC_SSE4))   have_e_sse4_1  = 0;
-
     if (!do_simd) { // SIMD disabled
         return order & 1
             ? rans_compress_O1_4x16
@@ -925,6 +977,14 @@ unsigned char *(*rans_enc_func(int do_simd, int order))
     }
 #endif
 
+    int have_e_sse4_1  = have_sse4_1;
+    int have_e_avx2    = have_avx2;
+    int have_e_avx512f = have_avx512f;
+
+    if (!(rans_cpu & RANS_CPU_ENC_AVX512)) have_e_avx512f = 0;
+    if (!(rans_cpu & RANS_CPU_ENC_AVX2))   have_e_avx2    = 0;
+    if (!(rans_cpu & RANS_CPU_ENC_SSE4))   have_e_sse4_1  = 0;
+
     if (order & 1) {
         // With simulated gathers, the AVX512 is now slower than AVX2, so
         // we avoid using it unless asking for the real avx512 gather.
@@ -974,14 +1034,6 @@ unsigned char *(*rans_dec_func(int do_simd, int order))
      unsigned char *out,
      unsigned int out_size) {
 
-    int have_d_sse4_1  = have_sse4_1;
-    int have_d_avx2    = have_avx2;
-    int have_d_avx512f = have_avx512f;
-
-    if (!(rans_cpu & RANS_CPU_DEC_AVX512)) have_d_avx512f = 0;
-    if (!(rans_cpu & RANS_CPU_DEC_AVX2))   have_d_avx2    = 0;
-    if (!(rans_cpu & RANS_CPU_DEC_SSE4))   have_d_sse4_1  = 0;
-
     if (!do_simd) { // SIMD disabled
         return order & 1
             ? rans_uncompress_O1_4x16
@@ -999,6 +1051,14 @@ unsigned char *(*rans_dec_func(int do_simd, int order))
     }
 #endif
 
+    int have_d_sse4_1  = have_sse4_1;
+    int have_d_avx2    = have_avx2;
+    int have_d_avx512f = have_avx512f;
+
+    if (!(rans_cpu & RANS_CPU_DEC_AVX512)) have_d_avx512f = 0;
+    if (!(rans_cpu & RANS_CPU_DEC_AVX2))   have_d_avx2    = 0;
+    if (!(rans_cpu & RANS_CPU_DEC_SSE4))   have_d_sse4_1  = 0;
+
     if (order & 1) {
 #if defined(HAVE_AVX512)
         if (have_d_avx512f)
@@ -1169,6 +1229,12 @@ unsigned char *rans_compress_to_4x16(unsigned char *in, unsigned int in_size,
         return NULL;
     }
 
+#ifdef VALIDATE_RANS
+    int orig_order = order;
+    int orig_in_size = in_size;
+    unsigned char *orig_in = in;
+#endif
+
     unsigned int c_meta_len;
     uint8_t *meta = NULL, *rle = NULL, *packed = NULL;
     uint8_t *out_free = NULL;
@@ -1385,6 +1451,11 @@ unsigned char *rans_compress_to_4x16(unsigned char *in, unsigned int in_size,
             int sz = var_put_u32(out+c_meta_len, out_end, in_size);
             c_meta_len += sz;
             *out_size -= sz;
+
+            if (do_simd && in_size < 32) {
+                do_simd = 0;
+                out[0] &= ~RANS_ORDER_X32;
+            }
         }
     } else if (do_pack) {
         out[0] &= ~RANS_ORDER_PACK;
@@ -1430,6 +1501,10 @@ unsigned char *rans_compress_to_4x16(unsigned char *in, unsigned int in_size,
                 return NULL;
             }
             c_rmeta_len = *out_size - (c_meta_len+sz+5);
+            if (do_simd && (rmeta_len < 32 || rle_len < 32)) {
+                do_simd = 0;
+                out[0] &= ~RANS_ORDER_X32;
+            }
             if (!rans_enc_func(do_simd, 0)(meta, rmeta_len, out+c_meta_len+sz+5, &c_rmeta_len)) {
                 free(out_free);
                 free(rle);
@@ -1503,6 +1578,24 @@ unsigned char *rans_compress_to_4x16(unsigned char *in, unsigned int in_size,
 
     *out_size += c_meta_len;
 
+// Validation mode
+#ifdef VALIDATE_RANS
+    unsigned int decoded_size = orig_in_size;
+    unsigned char *decoded = malloc(decoded_size);
+    decoded = rans_uncompress_to_4x16(out, *out_size,
+                                      decoded, &decoded_size);
+    if (!decoded ||
+        decoded_size != orig_in_size ||
+        memcmp(orig_in, decoded, orig_in_size) != 0) {
+        fprintf(stderr, "rans round trip failed for order %d. Written to fd 5\n", orig_order);
+        if (write(5, orig_in, orig_in_size) < 0)
+            abort();
+        abort();
+    }
+    free(decoded);
+#endif
+
+
     return out;
 }
 
