@@ -1,5 +1,5 @@
-/* 
-    Copyright (C) 2016 Genome Research Ltd.
+/*
+    Copyright (C) 2016-2025 Genome Research Ltd.
 
     Author: Petr Danecek <pd3@sanger.ac.uk>
 
@@ -9,10 +9,10 @@
     to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
     copies of the Software, and to permit persons to whom the Software is
     furnished to do so, subject to the following conditions:
-    
+
     The above copyright notice and this permission notice shall be included in
     all copies or substantial portions of the Software.
-    
+
     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
     IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
     FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
@@ -24,6 +24,7 @@
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <strings.h>
 #include <getopt.h>
 #include <stdarg.h>
 #include <stdint.h>
@@ -33,6 +34,11 @@
 #include <inttypes.h>
 #include <unistd.h>
 #include "bcftools.h"
+#include "filter.h"
+
+// Logic of the filters: include or exclude sites which match the filters?
+#define FLT_INCLUDE 1
+#define FLT_EXCLUDE 2
 
 #define GUESS_GT 1
 #define GUESS_PL 2
@@ -59,6 +65,10 @@ typedef struct
     char **argv, *af_tag;
     double af_dflt;
     stats_t stats;
+    filter_t *filter;
+    char *filter_str;
+    int filter_logic;   // include or exclude sites which match the filters? One of FLT_INCLUDE/FLT_EXCLUDE
+    const uint8_t *smpl_pass;
     int nsample, verbose, tag, include_indels;
     int *counts, ncounts;       // number of observed GTs with given ploidy, used when -g is not given
     double *tmpf, *pl2p, gt_err_prob;
@@ -78,10 +88,10 @@ const char *about(void)
 
 static const char *usage_text(void)
 {
-    return 
+    return
         "\n"
         "About: Determine sample sex by checking genotype likelihoods (GL,PL) or genotypes (GT)\n"
-        "       in the non-PAR region of chrX. The HWE is assumed, so given the alternate allele\n" 
+        "       in the non-PAR region of chrX. The HWE is assumed, so given the alternate allele\n"
         "       frequency fA and the genotype likelihoods pRR,pRA,pAA, the probabilities are\n"
         "       calculated as\n"
         "           P(dip) = pRR*(1-fA)^2 + pAA*fA^2 + 2*pRA*(1-fA)*fA\n"
@@ -94,19 +104,21 @@ static const char *usage_text(void)
         "\n"
         "Usage: bcftools +guess-ploidy <file.vcf.gz> [Plugin Options]\n"
         "Plugin options:\n"
-        "       --AF-dflt <float>           the default alternate allele frequency [0.5]\n"
-        "       --AF-tag <TAG>              use TAG for allele frequency\n"
-        "   -e, --error-rate <float>        probability of GT being wrong (with -t GT) [1e-3]\n"
-        "   -i, --include-indels            do not skip indel sites\n"
-        "   -g, --genome <str>              shortcut to select nonPAR region for common genomes b37|hg19|b38|hg38\n"
-        "   -r, --regions <chr:beg-end>     restrict to comma-separated list of regions\n"
-        "   -R, --regions-file <file>       restrict to regions listed in a file\n"
-        "   -t, --tag <tag>                 genotype or genotype likelihoods: GT, PL, GL [PL]\n"
-        "   -v, --verbose                   verbose output (specify twice to increase verbosity)\n"
+        "       --AF-dflt FLOAT             The default alternate allele frequency [0.5]\n"
+        "       --AF-tag TAG                Use TAG for allele frequency\n"
+        "   -e, --error-rate FLOAT          Probability of GT being wrong (with -t GT) [1e-3]\n"
+        "       --exclude EXPR              Exclude sites for which the expression is true\n"
+        "   -i, --include-indels            Do not skip indel sites\n"
+        "       --include EXPR              Include only sites for which the expression is true\n"
+        "   -g, --genome STR                Shortcut to select nonPAR region for common genomes b37|hg19|b38|hg38\n"
+        "   -r, --regions CHR:BEG-END       Restrict to comma-separated list of regions\n"
+        "   -R, --regions-file FILE         Restrict to regions listed in a file\n"
+        "   -t, --tag TAG                   Genotype or genotype likelihoods: GT, PL, GL [PL]\n"
+        "   -v, --verbosity INT             Verbose output (specify twice to increase verbosity)\n"
         "\n"
         "Region shortcuts:\n"
-        "   b37 .. -r X:2699521-154931043   # GRCh37 no-chr prefix\n"
-        "   b38 .. -r X:2781480-155701381   # GRCh38 no-chr prefix\n"
+        "   b37  .. -r X:2699521-154931043      # GRCh37 no-chr prefix\n"
+        "   b38  .. -r X:2781480-155701381      # GRCh38 no-chr prefix\n"
         "   hg19 .. -r chrX:2699521-154931043   # GRCh37 chr prefix\n"
         "   hg38 .. -r chrX:2781480-155701381   # GRCh38 chr prefix\n"
         "\n"
@@ -118,6 +130,15 @@ static const char *usage_text(void)
         "\n";
 }
 
+static inline int smpl_pass(args_t *args, int ismpl)
+{
+    if ( !args->smpl_pass ) return 1;
+    int pass = args->smpl_pass[ismpl];
+    if ( args->filter_logic & FLT_EXCLUDE ) pass = pass ? 0 : 1;
+    if ( pass ) return 1;
+    return 0;
+}
+
 void process_region_guess(args_t *args)
 {
     while ( bcf_sr_next_line(args->sr) )
@@ -126,6 +147,13 @@ void process_region_guess(args_t *args)
         if ( rec->n_allele==1 ) continue;
         if ( !args->include_indels && !(bcf_get_variant_types(rec)&VCF_SNP) ) continue;
 
+        if ( args->filter )
+        {
+            int pass = filter_test(args->filter, rec, &args->smpl_pass);
+            if ( args->filter_logic & FLT_EXCLUDE ) pass = pass ? 0 : 1;
+            if ( !args->smpl_pass && !pass ) continue;     // site-level filtering, not per-sample filtering
+        }
+
         double freq[2] = {0,0}, sum;
         int ismpl,i;
         if ( args->tag & GUESS_GT )   // use GTs to guess the ploidy, considering only one ALT
@@ -135,10 +163,11 @@ void process_region_guess(args_t *args)
             ngt /= args->nsample;
             for (ismpl=0; ismpl<args->nsample; ismpl++)
             {
+                if ( !smpl_pass(args,ismpl) ) continue;
                 int32_t *ptr = args->arr + ismpl*ngt;
                 double *tmp = args->tmpf + ismpl*3;
 
-                if ( ptr[0]==bcf_gt_missing ) 
+                if ( bcf_gt_is_missing(ptr[0]) )
                 {
                     tmp[0] = -1;
                     continue;
@@ -186,11 +215,12 @@ void process_region_guess(args_t *args)
             {
                 for (ismpl=0; ismpl<args->nsample; ismpl++)
                 {
+                    if ( !smpl_pass(args,ismpl) ) continue;
                     int32_t *ptr = args->arr + ismpl*npl;
                     double *tmp = args->tmpf + ismpl*3;
 
                     // restrict to first ALT
-                    if ( ptr[0]==bcf_int32_missing || ptr[1]==bcf_int32_missing || ptr[2]==bcf_int32_missing ) 
+                    if ( ptr[0]==bcf_int32_missing || ptr[1]==bcf_int32_missing || ptr[2]==bcf_int32_missing )
                     {
                         tmp[0] = -1;
                         continue;
@@ -230,11 +260,12 @@ void process_region_guess(args_t *args)
             {
                 for (ismpl=0; ismpl<args->nsample; ismpl++)
                 {
+                    if ( !smpl_pass(args,ismpl) ) continue;
                     int32_t *ptr = args->arr + ismpl*npl;
                     double *tmp = args->tmpf + ismpl*3;
 
                     // restrict to first ALT
-                    if ( ptr[0]==bcf_int32_missing || ptr[1]==bcf_int32_missing ) 
+                    if ( ptr[0]==bcf_int32_missing || ptr[1]==bcf_int32_missing )
                     {
                         tmp[0] = -1;
                         continue;
@@ -264,11 +295,12 @@ void process_region_guess(args_t *args)
             {
                 for (ismpl=0; ismpl<args->nsample; ismpl++)
                 {
+                    if ( !smpl_pass(args,ismpl) ) continue;
                     float *ptr = args->farr + ismpl*ngl;
                     double *tmp = args->tmpf + ismpl*3;
 
                     // restrict to first ALT
-                    if ( bcf_float_is_missing(ptr[0]) || bcf_float_is_missing(ptr[1]) || bcf_float_is_missing(ptr[2]) ) 
+                    if ( bcf_float_is_missing(ptr[0]) || bcf_float_is_missing(ptr[1]) || bcf_float_is_missing(ptr[2]) )
                     {
                         tmp[0] = -1;
                         continue;
@@ -308,11 +340,12 @@ void process_region_guess(args_t *args)
             {
                 for (ismpl=0; ismpl<args->nsample; ismpl++)
                 {
+                    if ( !smpl_pass(args,ismpl) ) continue;
                     float *ptr = args->farr + ismpl*ngl;
                     double *tmp = args->tmpf + ismpl*3;
 
                     // restrict to first ALT
-                    if ( bcf_float_is_missing(ptr[0]) || bcf_float_is_missing(ptr[1]) ) 
+                    if ( bcf_float_is_missing(ptr[0]) || bcf_float_is_missing(ptr[1]) )
                     {
                         tmp[0] = -1;
                         continue;
@@ -344,6 +377,7 @@ void process_region_guess(args_t *args)
         freq[1] /= sum;
         for (ismpl=0; ismpl<args->nsample; ismpl++)
         {
+            if ( !smpl_pass(args,ismpl) ) continue;
             count_t *counts = &args->stats.counts[ismpl];
             double *tmp = args->tmpf + ismpl*3;
             if ( tmp[0] < 0 ) continue;
@@ -353,7 +387,7 @@ void process_region_guess(args_t *args)
             counts->pdip += log(pdip);
             counts->ncount++;
             if ( args->verbose>1 )
-                printf("DBG\t%s\t%d\t%s\t%e\t%e\t%e\t%e\t%e\t%e\n", bcf_seqname(args->hdr,rec),rec->pos+1,bcf_hdr_int2id(args->hdr,BCF_DT_SAMPLE,ismpl),
+                printf("DBG\t%s\t%"PRId64"\t%s\t%e\t%e\t%e\t%e\t%e\t%e\n", bcf_seqname(args->hdr,rec),(int64_t) rec->pos+1,bcf_hdr_int2id(args->hdr,BCF_DT_SAMPLE,ismpl),
                     freq[1],tmp[0],tmp[1],tmp[2],phap,pdip);
         }
     }
@@ -372,7 +406,10 @@ int run(int argc, char **argv)
     {
         {"AF-tag",required_argument,NULL,0},
         {"AF-dflt",required_argument,NULL,1},
-        {"verbose",no_argument,NULL,'v'},
+        {"exclude",required_argument,NULL,2},
+        {"include",required_argument,NULL,3},
+        {"verbose",optional_argument,NULL,'v'},
+        {"verbosity",optional_argument,NULL,'v'},
         {"include-indels",no_argument,NULL,'i'},
         {"error-rate",required_argument,NULL,'e'},
         {"tag",required_argument,NULL,'t'},
@@ -384,15 +421,21 @@ int run(int argc, char **argv)
     };
     int c;
     char *tmp;
-    while ((c = getopt_long(argc, argv, "vr:R:t:e:ig:",loptions,NULL)) >= 0)
+    while ((c = getopt_long(argc, argv, "v::r:R:t:e:ig:",loptions,NULL)) >= 0)
     {
-        switch (c) 
+        switch (c)
         {
             case 0: args->af_tag = optarg; break;
-            case 1: 
+            case 1:
                     args->af_dflt = strtod(optarg,&tmp);
                     if ( *tmp ) error("Could not parse: --AF-dflt %s\n", optarg);
                     break;
+            case 2:
+                if ( args->filter_str ) error("Error: only one --include or --exclude expression can be given, and they cannot be combined\n");
+                args->filter_str = optarg; args->filter_logic |= FLT_EXCLUDE; break;
+            case 3:
+                if ( args->filter_str ) error("Error: only one --include or --exclude expression can be given, and they cannot be combined\n");
+                args->filter_str = optarg; args->filter_logic |= FLT_INCLUDE; break;
             case 'i': args->include_indels = 1; break;
             case 'e':
                 args->gt_err_prob = strtod(optarg,&tmp);
@@ -406,9 +449,17 @@ int run(int argc, char **argv)
                 else if ( !strcasecmp(optarg,"hg38") ) region = "chrX:2781480-155701381";
                 else error("The argument not recognised, expected --genome b37, b38, hg19 or hg38: %s\n", optarg);
                 break;
-            case 'R': region_is_file = 1; 
-            case 'r': region = optarg; break; 
-            case 'v': args->verbose++; break; 
+            case 'R': region_is_file = 1; // fall-through
+            case 'r': region = optarg; break;
+            case 'v':
+                if (!optarg) args->verbose++;
+                else
+                {
+                    args->verbose = strtol(optarg,&tmp,10);
+                    if ( *tmp || args->verbose<0 ) error("Could not parse argument: --verbosity %s\n", optarg);
+                    if ( args->verbose > 3 ) hts_verbose = args->verbose;
+                }
+                break;
             case 't':
                 if ( !strcasecmp(optarg,"GT") ) args->tag = GUESS_GT;
                 else if ( !strcasecmp(optarg,"PL") ) args->tag = GUESS_PL;
@@ -420,14 +471,15 @@ int run(int argc, char **argv)
             default: error("%s", usage_text()); break;
         }
     }
+    if ( args->filter_logic == (FLT_EXCLUDE|FLT_INCLUDE) ) error("Only one of --include or --exclude can be given.\n");
 
     char *fname = NULL;
     if ( optind==argc )
     {
         if ( !isatty(fileno((FILE *)stdin)) ) fname = "-";  // reading from stdin
-        else { error(usage_text()); }
+        else { error("%s",usage_text()); }
     }
-    else if ( optind+1!=argc ) error(usage_text());
+    else if ( optind+1!=argc ) error("%s",usage_text());
     else fname = argv[optind];
 
     args->sr = bcf_sr_init();
@@ -453,6 +505,9 @@ int run(int argc, char **argv)
     args->nsample = bcf_hdr_nsamples(args->hdr);
     args->stats.counts = (count_t*) calloc(args->nsample,sizeof(count_t));
 
+    if ( args->filter_str )
+        args->filter = filter_init(args->hdr, args->filter_str);
+
     if ( args->af_tag && !bcf_hdr_idinfo_exists(args->hdr,BCF_HL_INFO,bcf_hdr_id2int(args->hdr,BCF_DT_ID,args->af_tag)) )
         error("No such INFO tag: %s\n", args->af_tag);
 
@@ -509,7 +564,10 @@ int run(int argc, char **argv)
         else
             printf("%s\t%c\n", args->hdr->samples[i],predicted_sex);
     }
-   
+
+    if ( args->filter )
+        filter_destroy(args->filter);
+
     bcf_sr_destroy(args->sr);
     free(args->pl2p);
     free(args->tmpf);
