@@ -28,6 +28,7 @@
 #include <unistd.h>
 #include <ctype.h>
 #include <htslib/synced_bcf_reader.h>
+#include <htslib/hts_defs.h>
 #include "bcftools.h"
 #include "htslib/khash_str2int.h"
 #include "regidx.h"
@@ -97,7 +98,8 @@ static const char *usage_text(void)
         "   -x, --extra STRING              Output records not overlapping listed regions in separate file\n"
         "   -p, --prefix STRING             Prepend string to output VCF names\n"
         "       --hts-opts LIST             Low-level options to pass to HTSlib, e.g. block_size=32768\n"
-        "       --write-index               Automatically index the output files [off]\n"
+        "   -v, --verbosity INT             Verbosity level\n"
+        "   -W, --write-index[=FMT]         Automatically index the output files [off]\n"
         "\n"
         "Examples:\n"
         "   # Scatter a VCF file by shards with 10000 variants each\n"
@@ -108,7 +110,7 @@ static const char *usage_text(void)
         "\n";
 }
 
-void mkdir_p(const char *fmt, ...);
+void mkdir_p(const char *fmt, ...) HTS_FORMAT(HTS_PRINTF_FMT, 1, 2);
 
 // most of this code was inspired by Petr Danecek's code in regidx.c
 #define MAX_COOR_0 REGIDX_MAX   // CSI and hts_itr_query limit, 0-based
@@ -203,7 +205,9 @@ static void open_set(subset_t *set, args_t *args)
         if ( args->record_cmd_line ) bcf_hdr_append_version(args->hdr, args->argc, args->argv, "bcftools_plugin");
     }
     if ( bcf_hdr_write(set->fh, args->hdr)!=0 ) error("[%s] Error: cannot write the header to %s\n", __func__, args->str.s);
-    if ( args->write_index && init_index(set->fh,args->hdr,args->str.s,&set->index_fn)<0 ) error("Error: failed to initialise index for %s\n",args->str.s);
+    if ( init_index2(set->fh,args->hdr,args->str.s,&set->index_fn,
+                     args->write_index)<0 )
+        error("Error: failed to initialise index for %s\n",args->str.s);
 }
 
 static void init_data(args_t *args)
@@ -304,7 +308,12 @@ static void process(args_t *args)
         args->rec_cnt++;
         if (args->rec_cnt == args->nsites) {
           args->rec_cnt = 0;
+          int err = 0;
+          if ( args->write_index && bcf_idx_save(set->fh) < 0)
+              err = 1;
           if ( hts_close(set->fh)!=0 ) error("Error: close failed .. %s\n", set->fname);
+          if (err)
+              error("Error: cannot write to index %s\n", set->index_fn);
           free(set->fname);
           set->fname = NULL;
           args->chunk_cnt++;
@@ -352,15 +361,19 @@ int run(int argc, char **argv)
         {"extra",required_argument,NULL,'x'},
         {"prefix",required_argument,NULL,'p'},
         {"hts-opts",required_argument,NULL,5},
-        {"write-index",no_argument,NULL,6},
+        {"write-index",optional_argument,NULL,'W'},
+        {"verbosity",required_argument,NULL,'v'},
         {NULL,0,NULL,0}
     };
     int c;
     char *tmp;
-    while ((c = getopt_long(argc, argv, "e:i:o:O:r:R:t:T:n:s:S:x:p:h?", loptions, NULL)) >= 0)
+    while ((c = getopt_long(argc, argv, "e:i:o:O:r:R:t:T:n:s:S:x:p:W::h?v:", loptions, NULL)) >= 0)
     {
         switch (c)
         {
+            case 'v':
+                if ( apply_verbosity(optarg) < 0 ) error("Could not parse argument: --verbosity %s\n", optarg);
+                break;
             case 'e':
                 if ( args->filter_str ) error("Error: only one -i or -e expression can be given, and they cannot be combined\n");
                 args->filter_str = optarg; args->filter_logic |= FLT_EXCLUDE; break;
@@ -410,7 +423,10 @@ int run(int argc, char **argv)
             case 'x': args->extra = optarg;  break;
             case 'p': args->prefix = optarg;  break;
             case  5 : args->hts_opts = hts_readlist(optarg, 0, &args->nhts_opts); break;
-            case  6 : args->write_index = 1; break;
+            case 'W':
+                if (!(args->write_index = write_index_parse(optarg)))
+                    error("Unsupported index format '%s'\n", optarg);
+                break;
             case 'h':
             case '?':
             default: error("%s", usage_text()); break;
