@@ -1,6 +1,6 @@
 /*  plugins/setGT.c -- set gentoypes to given values
 
-    Copyright (C) 2015-2024 Genome Research Ltd.
+    Copyright (C) 2015-2025 Genome Research Ltd.
 
     Author: Petr Danecek <pd3@sanger.ac.uk>
 
@@ -82,10 +82,12 @@ args_t *args = NULL;
 #define GT_CUSTOM   (1<<10)
 #define GT_X_VAF    (1<<11)
 #define GT_RAND     (1<<12)
+#define GT_INVPHASE (1<<13)
 
-#define MINOR_ALLELE -1
-#define MAJOR_ALLELE -2
-#define X_VAF_ALLELE -3
+#define MINOR_ALLELE   -1
+#define MAJOR_ALLELE   -2
+#define X_VAF_ALLELE   -3
+#define MISSING_ALLELE -4
 
 const char *about(void)
 {
@@ -106,7 +108,8 @@ const char *usage(void)
         "       and the new genotype can be one of:\n"
         "           .       .. missing (\".\" or \"./.\", keeps ploidy)\n"
         "           0       .. reference allele (e.g. 0/0 or 0, keeps ploidy)\n"
-        "           c:GT    .. custom genotype (e.g. 0/0, 0, 0/1, m/M, 0/X overrides ploidy)\n"
+        "           c:GT    .. custom genotype (e.g. 0/0, 0, 0/1, m/M, 0/X, ./., .;  overrides ploidy)\n"
+        "           i       .. invert the genotype phase (0|1 becomes 1|0)\n"
         "           m       .. minor (the second most common) allele as determined from INFO/AC or FMT/GT (e.g. 1/1 or 1, keeps ploidy)\n"
         "           M       .. major allele as determined from INFO/AC or FMT/GT (e.g. 1/1 or 1, keeps ploidy)\n"
         "           X       .. allele with bigger read depth as determined from FMT/AD\n"
@@ -234,6 +237,7 @@ int init(int argc, char **argv, bcf_hdr_t *in, bcf_hdr_t *out)
                 if ( strchr(optarg,'X') ) args->new_mask |= GT_X_VAF;
                 if ( strchr(optarg,'p') ) args->new_mask |= GT_PHASED;
                 if ( strchr(optarg,'u') ) args->new_mask |= GT_UNPHASED;
+                if ( strchr(optarg,'i') ) args->new_mask |= GT_INVPHASE;
                 if ( !strncmp(optarg,"c:",2) ) { args->new_mask |= GT_CUSTOM; args->custom.gt_str = optarg; }
                 if ( args->new_mask==0 ) error("Unknown parameter to --new-gt: %s\n", optarg);
                 break;
@@ -279,6 +283,7 @@ int init(int argc, char **argv, bcf_hdr_t *in, bcf_hdr_t *out)
             if ( *ptr=='m' ) { allele = MINOR_ALLELE; args->new_mask |= GT_MINOR; }
             else if ( *ptr=='M' ) { allele = MAJOR_ALLELE; args->new_mask |= GT_MAJOR; }
             else if ( *ptr=='X' ) { allele = X_VAF_ALLELE; args->new_mask |= GT_X_VAF; }
+            else if ( *ptr=='.' ) allele = MISSING_ALLELE;
             else if ( *ptr=='/' || *ptr=='|' )
             {
                 if ( !args->custom.ploidy ) error("Could not parse the genotype: %s\n",args->custom.gt_str);
@@ -362,6 +367,21 @@ static inline int unphase_gt(int32_t *ptr, int ngts)
     return changed;
 }
 
+// invert the phase for a single sample, ngts is the ploidy
+static inline int invert_phase_gt(int32_t *ptr, int ngts)
+{
+    int gt;
+    if ( ngts!=2 ) return 0;    // don't invert the phase for sample ploidy != 2
+    if ( ptr[0]==bcf_int32_vector_end || ptr[1]==bcf_int32_vector_end ) return 0;
+    gt = bcf_gt_allele(ptr[1]);
+    if (bcf_gt_is_phased(ptr[1]))
+        ptr[1] = bcf_gt_phased(bcf_gt_allele(ptr[0]));
+    else
+        ptr[1] = bcf_gt_unphased(bcf_gt_allele(ptr[0]));
+    ptr[0] = bcf_gt_unphased(gt);    // first gt doesn't carry the phased flag per BCF specifications
+    return 2;
+}
+
 // sets GT for a single sample, ngts is the ploidy, allele
 static inline int set_gt(int32_t *ptr, int ngts, int allele)
 {
@@ -383,7 +403,8 @@ static inline int set_gt_custom(args_t *args, int32_t *ptr, int ngts, int nals)
     {
         if ( args->custom.gt[i]==MINOR_ALLELE ) new_allele = args->custom.m_allele;
         else if ( args->custom.gt[i]==MAJOR_ALLELE ) new_allele = args->custom.M_allele;
-        else if ( args->custom.gt[i]==X_VAF_ALLELE ) new_allele = args->custom.x_vaf_allele==bcf_gt_missing ? nals : bcf_gt_allele(args->custom.x_vaf_allele);
+        else if ( args->custom.gt[i]==X_VAF_ALLELE ) new_allele = bcf_gt_is_missing(args->custom.x_vaf_allele) ? nals : bcf_gt_allele(args->custom.x_vaf_allele);
+        else if ( args->custom.gt[i]==MISSING_ALLELE ) new_allele = nals; // this is to trigger the `new_allele = bcf_gt_missing` branch below
         else new_allele = args->custom.gt[i];
         if ( new_allele >= nals ) // cannot set, the requested index is bigger than there are alleles in ALT
             new_allele = bcf_gt_missing;
@@ -540,6 +561,8 @@ bcf1_t *process(bcf1_t *rec)
             }
             else if ( args->new_mask & GT_X_VAF )
                 changed += set_gt(ptr, ngts, args->xarr[i]);
+            else if ( args->new_mask & GT_INVPHASE )
+                changed += invert_phase_gt(ptr, ngts);
             else
                 changed += set_gt(ptr, ngts, args->new_gt);
         }
@@ -581,6 +604,8 @@ bcf1_t *process(bcf1_t *rec)
             }
             else if ( args->new_mask & GT_X_VAF )
                 changed += set_gt(args->gts + i*ngts, ngts, args->xarr[i]);
+            else if ( args->new_mask & GT_INVPHASE )
+                changed += invert_phase_gt(args->gts, ngts);
             else
                 changed += set_gt(args->gts + i*ngts, ngts, args->new_gt);
         }
@@ -617,6 +642,8 @@ bcf1_t *process(bcf1_t *rec)
             }
             else if ( args->new_mask & GT_X_VAF )
                 changed += set_gt(ptr, ngts, args->xarr[i]);
+            else if ( args->new_mask & GT_INVPHASE )
+                changed += invert_phase_gt(ptr, ngts);
             else
                 changed += set_gt(ptr, ngts, args->new_gt);
         }
