@@ -1,6 +1,6 @@
 /*  vcfplugin.c -- plugin modules for operating on VCF/BCF files.
 
-    Copyright (C) 2013-2021 Genome Research Ltd.
+    Copyright (C) 2013-2025 Genome Research Ltd.
 
     Author: Petr Danecek <pd3@sanger.ac.uk>
 
@@ -66,7 +66,7 @@ typedef struct _plugin_t plugin_t;
  *      success or non-zero value on error.
  *
  *   int init(int argc, char **argv, bcf_hdr_t *in_hdr, bcf_hdr_t *out_hdr)
- *      - called once at startup, allows to initialize local variables.
+ *      - called once at startup, it initializes local variables.
  *      Return 1 to suppress normal VCF/BCF header output, -1 on critical
  *      errors, 0 otherwise.
  *
@@ -149,6 +149,8 @@ typedef struct _args_t
 
     char **argv, *output_fname, *regions_list, *targets_list;
     int argc, drop_header, verbose, record_cmd_line, plist_only;
+    char *index_fn;
+    int write_index;
 }
 args_t;
 
@@ -455,11 +457,22 @@ static int cmp_plugin_name(const void *p1, const void *p2)
     return strcmp(a->name,b->name);
 }
 
+// If args=NULL then returns the number of plugins available. Otherwise prints the
+// plugins on stdout and returns 0 on success.
 static int list_plugins(args_t *args)
 {
     plugin_t *plugins = NULL;
     int nplugins = 0, mplugins = 0;
 
+    int count_only = 0;
+    args_t _args;
+    if ( !args )
+    {
+        memset(&_args,0,sizeof(_args));
+        args = &_args;
+        args->nplugin_paths = -1;
+        count_only = 1;
+    }
     init_plugin_paths(args);
 
     kstring_t str = {0,0,0};
@@ -490,6 +503,11 @@ static int list_plugins(args_t *args)
         }
         closedir(dp);
     }
+    if ( count_only )
+    {
+        free(str.s);
+        return nplugins;
+    }
     if ( nplugins )
     {
         qsort(plugins, nplugins, sizeof(plugins[0]), cmp_plugin_name);
@@ -508,6 +526,10 @@ static int list_plugins(args_t *args)
     free(str.s);
     return nplugins ? 0 : 1;
 }
+int count_plugins(void)
+{
+    return list_plugins(NULL);
+}
 
 static void init_data(args_t *args)
 {
@@ -528,6 +550,9 @@ static void init_data(args_t *args)
         if ( args->out_fh == NULL ) error("Can't write to \"%s\": %s\n", args->output_fname, strerror(errno));
         if ( args->n_threads ) hts_set_threads(args->out_fh, args->n_threads);
         if ( bcf_hdr_write(args->out_fh, args->hdr_out)!=0 ) error("[%s] Error: cannot write to %s\n", __func__,args->output_fname);
+        if ( init_index2(args->out_fh,args->hdr_out,args->output_fname,
+                         &args->index_fn, args->write_index)<0 )
+            error("Error: failed to initialise index for %s\n",args->output_fname);
     }
 }
 
@@ -549,7 +574,19 @@ static void destroy_data(args_t *args)
     }
     if ( args->filter )
         filter_destroy(args->filter);
-    if (args->out_fh && hts_close(args->out_fh)!=0 ) error("[%s] Error: close failed .. %s\n", __func__,args->output_fname);
+    if (args->out_fh )
+    {
+        if ( args->write_index )
+        {
+            if ( bcf_idx_save(args->out_fh)<0 )
+            {
+                if ( hts_close(args->out_fh)!=0 ) error("Error: close failed .. %s\n", args->output_fname?args->output_fname:"stdout");
+                error("Error: cannot write to index %s\n", args->index_fn);
+            }
+            free(args->index_fn);
+        }
+        if ( hts_close(args->out_fh)!=0 ) error("[%s] Error: close failed .. %s\n", __func__,args->output_fname);
+    }
 }
 
 static void usage(args_t *args)
@@ -564,10 +601,10 @@ static void usage(args_t *args)
     fprintf(stderr, "   -i, --include EXPR             Select sites for which the expression is true\n");
     fprintf(stderr, "   -r, --regions REGION           Restrict to comma-separated list of regions\n");
     fprintf(stderr, "   -R, --regions-file FILE        Restrict to regions listed in a file\n");
-    fprintf(stderr, "        --regions-overlap 0|1|2   Include if POS in the region (0), record overlaps (1), variant overlaps (2) [1]\n");
+    fprintf(stderr, "       --regions-overlap 0|1|2    Include if POS in the region (0), record overlaps (1), variant overlaps (2) [1]\n");
     fprintf(stderr, "   -t, --targets REGION           Similar to -r but streams rather than index-jumps\n");
     fprintf(stderr, "   -T, --targets-file FILE        Similar to -R but streams rather than index-jumps\n");
-    fprintf(stderr, "        --targets-overlap 0|1|2   Include if POS in the region (0), record overlaps (1), variant overlaps (2) [0]\n");
+    fprintf(stderr, "       --targets-overlap 0|1|2    Include if POS in the region (0), record overlaps (1), variant overlaps (2) [0]\n");
     fprintf(stderr, "VCF output options:\n");
     fprintf(stderr, "       --no-version               Do not append version and command line to the header\n");
     fprintf(stderr, "   -o, --output FILE              Write output to a file [standard output]\n");
@@ -576,8 +613,9 @@ static void usage(args_t *args)
     fprintf(stderr, "Plugin options:\n");
     fprintf(stderr, "   -h, --help                     List plugin's options\n");
     fprintf(stderr, "   -l, --list-plugins             List available plugins. See BCFTOOLS_PLUGINS environment variable and man page for details\n");
-    fprintf(stderr, "   -v, --verbose                  Print verbose information, -vv increases verbosity\n");
+    fprintf(stderr, "   -v, --verbosity INT            Verbosity level\n");
     fprintf(stderr, "   -V, --version                  Print version string and exit\n");
+    fprintf(stderr, "   -W, --write-index[=FMT]        Automatically index the output files [off]\n");
     fprintf(stderr, "\n");
     exit(1);
 }
@@ -623,9 +661,9 @@ int main_plugin(int argc, char *argv[])
     if ( argv[1][0]!='-' )
     {
         args->verbose = is_verbose(argc, argv);
-        plugin_name = argv[1]; 
-        argc--; 
-        argv++; 
+        plugin_name = argv[1];
+        argc--;
+        argv++;
         load_plugin(args, plugin_name, 1, &args->plugin);
         if ( args->plugin.run )
         {
@@ -640,7 +678,8 @@ int main_plugin(int argc, char *argv[])
     static struct option loptions[] =
     {
         {"version",no_argument,NULL,'V'},
-        {"verbose",no_argument,NULL,'v'},
+        {"verbose",optional_argument,NULL,'v'},
+        {"verbosity",optional_argument,NULL,'v'},
         {"help",no_argument,NULL,'h'},
         {"list-plugins",no_argument,NULL,'l'},
         {"output",required_argument,NULL,'o'},
@@ -655,14 +694,22 @@ int main_plugin(int argc, char *argv[])
         {"targets-file",required_argument,NULL,'T'},
         {"targets-overlap",required_argument,NULL,2},
         {"no-version",no_argument,NULL,8},
+        {"write-index",optional_argument,NULL,'W'},
         {NULL,0,NULL,0}
     };
     char *tmp;
-    while ((c = getopt_long(argc, argv, "h?o:O:r:R:t:T:li:e:vV",loptions,NULL)) >= 0)
+    while ((c = getopt_long(argc, argv, "h?o:O:r:R:t:T:li:e:v::VW::",loptions,NULL)) >= 0)
     {
         switch (c) {
             case 'V': version_only = 1; break;
-            case 'v': args->verbose++; break;
+            case 'v':
+                if ( !optarg ) args->verbose++;
+                else
+                {
+                    args->verbose = strtol(optarg,&tmp,10);
+                    if ( *tmp || args->verbose<0 ) error("Could not parse argument: --verbosity %s\n", optarg);
+                    if ( args->verbose > 3 ) hts_verbose = args->verbose;
+                }
             case 'o': args->output_fname = optarg; break;
             case 'O':
                 switch (optarg[0]) {
@@ -694,19 +741,19 @@ int main_plugin(int argc, char *argv[])
             case 'T': args->targets_list = optarg; targets_is_file = 1; break;
             case 'l': args->plist_only = 1; break;
             case  1 :
-                if ( !strcasecmp(optarg,"0") ) regions_overlap = 0;
-                else if ( !strcasecmp(optarg,"1") ) regions_overlap = 1;
-                else if ( !strcasecmp(optarg,"2") ) regions_overlap = 2;
-                else error("Could not parse: --regions-overlap %s\n",optarg);
+                regions_overlap = parse_overlap_option(optarg);
+                if ( regions_overlap < 0 ) error("Could not parse: --regions-overlap %s\n",optarg);
                 break;
             case  2 :
-                if ( !strcasecmp(optarg,"0") ) targets_overlap = 0;
-                else if ( !strcasecmp(optarg,"1") ) targets_overlap = 1;
-                else if ( !strcasecmp(optarg,"2") ) targets_overlap = 2;
-                else error("Could not parse: --targets-overlap %s\n",optarg);
+                targets_overlap = parse_overlap_option(optarg);
+                if ( targets_overlap < 0 ) error("Could not parse: --targets-overlap %s\n",optarg);
                 break;
             case  9 : args->n_threads = strtol(optarg, 0, 0); break;
             case  8 : args->record_cmd_line = 0; break;
+            case 'W':
+                if (!(args->write_index = write_index_parse(optarg)))
+                    error("Unsupported index format '%s'\n", optarg);
+                break;
             case '?':
             case 'h': usage_only = 1; break;
             default: error("Unknown argument: %s\n", optarg);
